""" This file is part of the Gudhi Library - https://gudhi.inria.fr/ - which is released under MIT.
    See file LICENSE or go to https://gudhi.inria.fr/licensing/ for full license details.
    Author(s):       Mathieu Carriere

    Copyright (C) 2020 Inria

    Modification(s):
      - YYYY/MM Author: Description of the modification
"""

from gudhi.wasserstein import wasserstein_distance as pot
import numpy as np
import torch
import tensorflow as tf

def test_wasserstein_distance_grad():
    diag1 = torch.tensor([[2.7, 3.7], [9.6, 14.0], [34.2, 34.974]], requires_grad=True)
    diag2 = torch.tensor([[2.8, 4.45], [9.5, 14.1]], requires_grad=True)
    diag3 = torch.tensor([[2.8, 4.45], [9.5, 14.1]], requires_grad=True)
    assert diag1.grad is None and diag2.grad is None and diag3.grad is None
    dist12 = pot(diag1, diag2, internal_p=2, order=2, enable_autodiff=True)
    dist30 = pot(diag3, torch.tensor([]), internal_p=2, order=2, enable_autodiff=True)
    dist12.backward()
    dist30.backward()
    assert not torch.isnan(diag1.grad).any() and not torch.isnan(diag2.grad).any() and not torch.isnan(diag3.grad).any()
    diag4 = torch.tensor([[0., 10.]], requires_grad=True)
    diag5 = torch.tensor([[1., 11.], [3., 4.]], requires_grad=True)
    dist45 = pot(diag4, diag5, internal_p=1, order=1, enable_autodiff=True)
    assert dist45 == 3.
    dist45.backward()
    assert np.array_equal(diag4.grad, [[-1., -1.]])
    assert np.array_equal(diag5.grad, [[1., 1.], [-1., 1.]])
    diag6 = torch.tensor([[5., 10.]], requires_grad=True)
    pot(diag6, diag6, internal_p=2, order=2, enable_autodiff=True).backward()
    # https://github.com/jonasrauber/eagerpy/issues/6
    # assert np.array_equal(diag6.grad, [[0., 0.]])

def test_wasserstein_distance_grad_tensorflow():
    with tf.GradientTape() as tape:
        diag4 = tf.convert_to_tensor(tf.Variable(initial_value=np.array([[0., 10.]]),           trainable=True))
        diag5 = tf.convert_to_tensor(tf.Variable(initial_value=np.array([[1., 11.], [3., 4.]]), trainable=True))
        dist45 = pot(diag4, diag5, internal_p=1, order=1, enable_autodiff=True)
        assert dist45 == 3.

    grads = tape.gradient(dist45, [diag4, diag5])
    assert np.array_equal(grads[0].values, [[-1., -1.]])
    assert np.array_equal(grads[1].values, [[1., 1.], [-1., 1.]])


# This file is part of the Gudhi Library - https://gudhi.inria.fr/ - which is released under MIT.
# See file LICENSE or go to https://gudhi.inria.fr/licensing/ for full license details.
# Author(s):       Marc Glisse
#
# Copyright (C) 2020 Inria
#
# Modification(s):
#   - YYYY/MM Author: Description of the modification

from .knn import KNearestNeighbors
import numpy as np

__author__ = "Marc Glisse"
__copyright__ = "Copyright (C) 2020 Inria"
__license__ = "MIT"


class DistanceToMeasure:
    """
    Class to compute the distance to the empirical measure defined by a point set, as introduced in :cite:`dtm`.
    """

    def __init__(self, k, q=2, **kwargs):
        """
        Args:
            k (int): number of neighbors (possibly including the point itself).
            q (float): order used to compute the distance to measure. Defaults to 2.
            kwargs: same parameters as :class:`~gudhi.point_cloud.knn.KNearestNeighbors`, except that
                metric="neighbors" means that :func:`transform` expects an array with the distances
                to the k nearest neighbors.
        """
        self.k = k
        self.q = q
        self.params = kwargs

    def fit_transform(self, X, y=None):
        return self.fit(X).transform(X)

    def fit(self, X, y=None):
        """
        Args:
            X (numpy.array): coordinates for mass points.
        """
        if self.params.setdefault("metric", "euclidean") != "neighbors":
            self.knn = KNearestNeighbors(
                self.k, return_index=False, return_distance=True, sort_results=False, **self.params
            )
            self.knn.fit(X)
        return self

    def transform(self, X):
        """
        Args:
            X (numpy.array): coordinates for query points, or distance matrix if metric is "precomputed",
                or distances to the k nearest neighbors if metric is "neighbors" (if the array has more
                than k columns, the remaining ones are ignored).

        Returns:
            numpy.array: a 1-d array with, for each point of X, its distance to the measure defined
            by the argument of :func:`fit`.
        """
        if self.params["metric"] == "neighbors":
            distances = X[:, : self.k]
        else:
            distances = self.knn.transform(X)
        distances = distances ** self.q
        dtm = distances.sum(-1) / self.k
        dtm = dtm ** (1.0 / self.q)
        # We compute too many powers, 1/p in knn then q in dtm, 1/q in dtm then q or some log in the caller.
        # Add option to skip the final root?
        return dtm


class DTMDensity:
    """
    Density estimator based on the distance to the empirical measure defined by a point set, as defined
    in :cite:`dtmdensity`. Note that this implementation only renormalizes when asked, and the renormalization
    only works for a Euclidean metric, so in other cases the total measure may not be 1.

    .. note:: When the dimension is high, using it as an exponent can quickly lead to under- or overflows.
        We recommend using a small fixed value instead in those cases, even if it won't have the same nice
        theoretical properties as the dimension.
    """

    def __init__(self, k=None, weights=None, q=None, dim=None, normalize=False, n_samples=None, **kwargs):
        """
        Args:
            k (int): number of neighbors (possibly including the point itself). Optional if it can be guessed
                from weights or metric="neighbors".
            weights (numpy.array): weights of each of the k neighbors, optional. They are supposed to sum to 1.
            q (float): order used to compute the distance to measure. Defaults to dim.
            dim (float): final exponent representing the dimension. Defaults to the dimension, and must be specified
                when the dimension cannot be read from the input (metric is "neighbors" or "precomputed").
            normalize (bool): normalize the density so it corresponds to a probability measure on ℝᵈ.
                Only available for the Euclidean metric, defaults to False.
            n_samples (int): number of sample points used for fitting. Only needed if `normalize` is True and
                metric is "neighbors".
            kwargs: same parameters as :class:`~gudhi.point_cloud.knn.KNearestNeighbors`, except that
                metric="neighbors" means that :func:`transform` expects an array with the distances to
                the k nearest neighbors.
        """
        if weights is None:
            self.k = k
            if k is None:
                assert kwargs.get("metric") == "neighbors", 'Must specify k or weights, unless metric is "neighbors"'
                self.weights = None
            else:
                self.weights = np.full(k, 1.0 / k)
        else:
            self.weights = weights
            self.k = len(weights)
            assert k is None or k == self.k, "k differs from the length of weights"
        self.q = q
        self.dim = dim
        self.params = kwargs
        self.normalize = normalize
        self.n_samples = n_samples

    def fit_transform(self, X, y=None):
        return self.fit(X).transform(X)

    def fit(self, X, y=None):
        """
        Args:
            X (numpy.array): coordinates for mass points.
        """
        if self.params.setdefault("metric", "euclidean") != "neighbors":
            self.knn = KNearestNeighbors(
                self.k, return_index=False, return_distance=True, sort_results=False, **self.params
            )
            self.knn.fit(X)
            if self.params["metric"] != "precomputed":
                self.n_samples = len(X)
        return self

    def transform(self, X):
        """
        Args:
            X (numpy.array): coordinates for query points, or distance matrix if metric is "precomputed",
                or distances to the k nearest neighbors if metric is "neighbors" (if the array has more
                than k columns, the remaining ones are ignored).
        """
        q = self.q
        dim = self.dim
        if dim is None:
            assert self.params["metric"] not in {
                "neighbors",
                "precomputed",
            }, "dim not specified and cannot guess the dimension"
            dim = len(X[0])
        if q is None:
            q = dim
        k = self.k
        weights = self.weights
        if self.params["metric"] == "neighbors":
            distances = np.asarray(X)
            if weights is None:
                k = distances.shape[1]
                weights = np.full(k, 1.0 / k)
            else:
                distances = distances[:, :k]
        else:
            distances = self.knn.transform(X)
        distances = distances ** q
        dtm = (distances * weights).sum(-1)
        if self.normalize:
            dtm /= (np.arange(1, k + 1) ** (q / dim) * weights).sum()
        density = dtm ** (-dim / q)
        if self.normalize:
            import math

            if self.params["metric"] == "precomputed":
                self.n_samples = len(X[0])
            # Volume of d-ball
            Vd = math.pi ** (dim / 2) / math.gamma(dim / 2 + 1)
            density /= self.n_samples * Vd
        return density
        # We compute too many powers, 1/p in knn then q in dtm, d/q in dtm then whatever in the caller.
        # Add option to skip the final root?



#!/usr/bin/env python

import km
import numpy as np
from collections import defaultdict
import argparse

"""This file is part of the Gudhi Library - https://gudhi.inria.fr/ - which is released under MIT.
    See file LICENSE or go to https://gudhi.inria.fr/licensing/ for full license details.
    Author(s):       Mathieu Carriere

   Copyright (C) 2017 Inria

   Modification(s):
     - YYYY/MM Author: Description of the modification
"""

__author__ = "Mathieu Carriere"
__copyright__ = "Copyright (C) 2017 Inria"
__license__ = "GPL v3"

parser = argparse.ArgumentParser(description='Creates an html Keppler Mapper '
                                 'file to visualize a SC.txt file.',
                                 epilog='Example: '
                                 './KeplerMapperVisuFromTxtFile.py '
                                 '-f ../../data/points/human.off_sc.txt'
                                 '- Constructs an human.off_sc.html file.')
parser.add_argument("-f", "--file", type=str, required=True)

args = parser.parse_args()

with open(args.file, 'r') as f:
    network = {}
    mapper = km.KeplerMapper(verbose=0)
    data = np.zeros((3,3))
    projected_data = mapper.fit_transform( data, projection="sum", scaler=None )

    nodes = defaultdict(list)
    links = defaultdict(list)
    custom = defaultdict(list)

    dat = f.readline()
    lens = f.readline()
    color = f.readline();
    param = [float(i) for i in f.readline().split(" ")]

    nums = [int(i) for i in f.readline().split(" ")]
    num_nodes = nums[0]
    num_edges = nums[1]

    for i in range(0,num_nodes):
        point = [float(j) for j in f.readline().split(" ")]
        nodes[  str(int(point[0]))  ] = [  int(point[0]), point[1], int(point[2])  ]
        links[  str(int(point[0]))  ] = []
        custom[  int(point[0])  ] = point[1]

    m = min([custom[i] for i in range(0,num_nodes)])
    M = max([custom[i] for i in range(0,num_nodes)])

    for i in range(0,num_edges):
        edge = [int(j) for j in f.readline().split(" ")]
        links[  str(edge[0])  ].append(  str(edge[1])  )
        links[  str(edge[1])  ].append(  str(edge[0])  )

    network["nodes"] = nodes
    network["links"] = links
    network["meta"] = lens

    html_output_filename = args.file.rsplit('.', 1)[0] + '.html'
    mapper.visualize(network, color_function = color, path_html=html_output_filename, title=dat,
    graph_link_distance=30, graph_gravity=0.1, graph_charge=-120, custom_tooltips=custom, width_html=0,
    height_html=0, show_tooltips=True, show_title=True, show_meta=True, res=param[0],gain=param[1], minimum=m,maximum=M)
    message = repr(html_output_filename) + " is generated. You can now use your favorite web browser to visualize it."
    print(message)


    f.close()



from gudhi.wasserstein.barycenter import lagrangian_barycenter
import numpy as np

""" This file is part of the Gudhi Library - https://gudhi.inria.fr/ - which is released under MIT.
    See file LICENSE or go to https://gudhi.inria.fr/licensing/ for full license details.
    Author(s):       Theo Lacombe

    Copyright (C) 2019 Inria

    Modification(s):
      - YYYY/MM Author: Description of the modification
"""

__author__ = "Theo Lacombe"
__copyright__ = "Copyright (C) 2019 Inria"
__license__ = "MIT"


def test_lagrangian_barycenter():

    dg1 = np.array([[0.2, 0.5]])
    dg2 = np.array([[0.2, 0.7]])
    dg3 = np.array([[0.3, 0.6], [0.7, 0.8], [0.2, 0.3]])
    dg4 = np.array([])
    dg5 = np.array([])
    dg6 = np.array([])
    res = np.array([[0.27916667, 0.55416667], [0.7375,  0.7625], [0.2375, 0.2625]])

    dg7 = np.array([[0.1, 0.15], [0.1, 0.7], [0.2, 0.22], [0.55, 0.84], [0.11, 0.91], [0.61, 0.75], [0.33, 0.46], [0.12, 0.41], [0.32, 0.48]])
    dg8 = np.array([[0., 4.], [4, 8]])

    # error crit.
    eps = 1e-7


    assert np.linalg.norm(lagrangian_barycenter(pdiagset=[dg1, dg2, dg3, dg4],init=3, verbose=False) - res) < eps
    assert np.array_equal(lagrangian_barycenter(pdiagset=[dg4, dg5, dg6], verbose=False), np.empty(shape=(0,2)))
    assert np.linalg.norm(lagrangian_barycenter(pdiagset=[dg7], verbose=False) - dg7) < eps
    Y, log = lagrangian_barycenter(pdiagset=[dg4, dg8], verbose=True)
    assert np.linalg.norm(Y - np.array([[1,3], [5, 7]])) < eps
    assert np.abs(log["energy"] - 2) < eps
    assert np.array_equal(log["groupings"][0] , np.array([[0, -1], [1, -1]]))
    assert np.array_equal(log["groupings"][1] , np.array([[0, 0], [1, 1]]))
    assert np.linalg.norm(lagrangian_barycenter(pdiagset=[dg8, dg4], init=np.array([[0.2, 0.6], [0.5, 0.7]]), verbose=False) - np.array([[1, 3], [5, 7]])) < eps
    assert lagrangian_barycenter(pdiagset = []) is None




from gudhi.point_cloud.timedelay import TimeDelayEmbedding
import numpy as np


def test_normal():
    # Sample array
    ts = [1, 2, 3, 4, 5, 6, 7, 8, 9, 10]
    # Normal case.
    prep = TimeDelayEmbedding()
    pointclouds = prep(ts)
    assert (pointclouds[0] == np.array([1, 2, 3])).all()
    assert (pointclouds[1] == np.array([2, 3, 4])).all()
    assert (pointclouds[2] == np.array([3, 4, 5])).all()
    assert (pointclouds[3] == np.array([4, 5, 6])).all()
    assert (pointclouds[4] == np.array([5, 6, 7])).all()
    assert (pointclouds[5] == np.array([6, 7, 8])).all()
    assert (pointclouds[6] == np.array([7, 8, 9])).all()
    assert (pointclouds[7] == np.array([8, 9, 10])).all()
    # Delay = 3
    prep = TimeDelayEmbedding(delay=3)
    pointclouds = prep(ts)
    assert (pointclouds[0] == np.array([1, 4, 7])).all()
    assert (pointclouds[1] == np.array([2, 5, 8])).all()
    assert (pointclouds[2] == np.array([3, 6, 9])).all()
    assert (pointclouds[3] == np.array([4, 7, 10])).all()
    # Skip = 3
    prep = TimeDelayEmbedding(skip=3)
    pointclouds = prep(ts)
    assert (pointclouds[0] == np.array([1, 2, 3])).all()
    assert (pointclouds[1] == np.array([4, 5, 6])).all()
    assert (pointclouds[2] == np.array([7, 8, 9])).all()
    # Delay = 2 / Skip = 2
    prep = TimeDelayEmbedding(delay=2, skip=2)
    pointclouds = prep(ts)
    assert (pointclouds[0] == np.array([1, 3, 5])).all()
    assert (pointclouds[1] == np.array([3, 5, 7])).all()
    assert (pointclouds[2] == np.array([5, 7, 9])).all()

    # Vector series
    ts = np.arange(0, 10).reshape(-1, 2)
    prep = TimeDelayEmbedding(dim=4)
    prep.fit([ts])
    assert (prep.transform([ts])[0] == [[0, 1, 2, 3], [2, 3, 4, 5], [4, 5, 6, 7], [6, 7, 8, 9]]).all()



# This file is part of the Gudhi Library - https://gudhi.inria.fr/ - which is released under MIT.
# See file LICENSE or go to https://gudhi.inria.fr/licensing/ for full license details.
# Author(s):       Yuichi Ike, Raphaël Tinarrage
#
# Copyright (C) 2020 Inria, Copyright (C) 2020 FUjitsu Laboratories Ltd.
#
# Modification(s):
#   - YYYY/MM Author: Description of the modification


from gudhi.weighted_rips_complex import WeightedRipsComplex
from gudhi.point_cloud.dtm import DistanceToMeasure
from scipy.spatial.distance import cdist

class DTMRipsComplex(WeightedRipsComplex):
    """
    Class to generate a DTM Rips complex from a distance matrix or a point set, 
    in the way described in :cite:`dtmfiltrations`.
    Remark that all the filtration values are doubled compared to the definition in the paper 
    for the consistency with RipsComplex.
    :Requires: `SciPy <installation.html#scipy>`_
    """
    def __init__(self, 
                 points=None, 
                 distance_matrix=None, 
                 k=1, 
                 q=2,
                 max_filtration=float('inf')):
        """
        Args:
            points (numpy.ndarray): array of points.
            distance_matrix (numpy.ndarray): full distance matrix.
            k (int): number of neighbors for the computation of DTM. Defaults to 1, which is equivalent to the usual Rips complex.
            q (float): order used to compute the distance to measure. Defaults to 2.
            max_filtration (float): specifies the maximal filtration value to be considered.      
        """
        if distance_matrix is None:
            if points is None:
                # Empty Rips construction
                points=[]
            distance_matrix = cdist(points,points)
        self.distance_matrix = distance_matrix

        # TODO: address the error when k is too large 
        if k <= 1:
            self.weights = [0] * len(distance_matrix)
        else:
            dtm = DistanceToMeasure(k, q=q, metric="precomputed")        
            self.weights = dtm.fit_transform(distance_matrix)
        self.max_filtration = max_filtration
        
